from pydoc import doc
import re
from lxml import etree
import lxml.html as H
from collections import defaultdict

tags = [
    'param',
'field',
'return',
'see',
'parent',
'content',
'firstline',
'inherit',
'generaltag',
]    
def parseDocstring(docstring) :
    '''
    Parse Docstring as returned by eclcc,
    break into individual tags and
    return them as XML Elements
    '''
    tags_re = "|".join(tags)
    docstring = re.sub(r'([@])\s+({0})'.format(tags_re), r'\1' + r'\2', docstring)

    docstring = re.sub(r'\n\s*\*', '\n', docstring)
    docstring = re.sub(r'\r', ' ', docstring)
    docstring = docstring.strip().split('\n')

    docdict = defaultdict(list)
    current_tag = 'content'
    current_text = ""
    for line in docstring :
        is_tag = re.search(r'^\s*@', line)
        if is_tag :
            if current_tag == 'content' :
                docdict['firstline'] = [findFirstLine(current_text)]
            docdict[current_tag].append(current_text.strip())
            line = re.split(r'\s', line.lstrip(), maxsplit=1)
            tag = line[0][1:]
            if len(line) > 1:
              text = line[1]
            else:
              text = ""
            current_tag = tag
            current_text = text + '\n'
        else :
            current_text += line + '\n'


    if current_tag == 'content' :
        docdict['firstline'] = [findFirstLine(current_text)]
    docdict[current_tag].append(current_text.strip())

    for tag in docdict :
        for i, desc in enumerate(docdict[tag]) :
            root = H.fragment_fromstring(desc, create_parent='div')
            removeWS(root)
            content = etree.Element(tag)
            content.text = etree.tostring(root)
            content.text = re.sub(r'^<div>', '', content.text)
            content.text = re.sub(r'</div>$', '', content.text)
            docdict[tag][i] = content

    return docdict

def removeWS(element) :
    '''
    Format Whitespace in HTML elements in docstring
    coming from parsed XML Output of ECL File
    '''
    if element.tag == 'pre' :
        lines = element.text.split('\n')
        element.text = lines[0]

        for line in lines[1:] :
            br = etree.Element('br')
            br.tail = line
            element.append(br)
        return

    if element.text is not None :
        element.text = re.sub(r'\s+', ' ', element.text)
    for e in element.iterchildren() :
        if e.tail :
            e.tail = re.sub(r'\s+', ' ', e.tail)

        removeWS(e)

def findFirstLine(current_text) :
    '''
    Find First line in docstring content section to be used as caption
    in TOC and Tree
    '''
    split_1 = re.split(r'\.\s|\.$', current_text.strip(), maxsplit=1)
    if len(split_1) == 2 :
        return split_1[0].strip()

    split_2 = re.split(r'\n', current_text.strip(), maxsplit=1)
    return split_2[0].strip()


def construct_type(ele) :
    '''
    Parse Type Tree into single string representation
    '''
    if ele is None : return ''
    if type(ele) == list : return ''

    typestring = ''
    attribs = ele.attrib
    typename = attribs['type']
    if typename == 'record' :
        if 'unnamed' in attribs :
            typestring += '{ '
            fields = []
            for field in ele.findall('Field') :
                fields.append(construct_type(field.find('./Type')) + " " + field.attrib['name'])
            typestring += ' , '.join(fields) + ' }'
        else :
            typestring += attribs['origfn'] if 'origfn' in attribs else attribs['name']
    else :
        typestring += typename.upper()
        if 'origfn' in attribs :
            typestring += ' ( ' + attribs['origfn'] + ' )'
        elif 'name' in attribs :
            typestring += ' ( ' + attribs['name'] + ' )'

    if typename == 'function' :
        typestring += ' [ '
        params = []
        for p in ele.find('Params').findall('Type') :
            params.append(construct_type(p))
        typestring += ' , '.join(params) + ' ]'

    if ele.find('./Type') is not None :
        typestring += ' ( ' + construct_type(ele.find('./Type')) + ' )'

    return typestring

def cleansign(text) :
    '''
    Remove irrelevant prefix and suffixes from signature
    '''
    print('cleansign in = "', text, '"')
    endpos = text.rfind(':=');
    if endpos >= 0:
      text = text[:endpos]
    text = re.sub(r'^export', '', text.strip(), flags=re.I)
    text = re.sub(r'^shared', '', text, flags=re.I)
    text = re.sub(r';$', '', text, flags=re.I)
    text = re.sub(r'\s+', ' ', text.strip())
    print('cleansign out = ', text)
    return text

def breaksign(name, text) :
    '''
    Heuristically break signature of ECL Definition
    recovered from ecl file into "return name (Paramters)"
    '''
    print('breaksign line = ', name, text)
    name = name.lower()
    string = ' ' + text.lower() + ' '
    pos = 1
    open_bracks = ['{', '(', '[']
    close_bracks = ['}', ')', ']']
    stack = []

    ret, param = '', ''
    indent_len = 0
    name_len = len(name)

    for i in range(1, len(string)) :
        c = string[i]
        if c in open_bracks :
            stack.append(c)
        elif c in close_bracks :
            if stack[-1] == open_bracks[close_bracks.index(c)] :
                stack = stack[:-1]
        else :
            if len(stack) == 0 :
                m = re.match(r'[\s\)]' + name + r'([^0-9A-Za-z_])', string[pos-1:])
                if m :
                    pos = pos - 1
                    ret = text[:pos]
                    param = text[pos + name_len:]
                    indent_len = pos + name_len
                    break

        pos += 1

    return ret.strip(), param.strip(), indent_len


def getTags(doc) :
    '''
    Convert XML Documentation (generated using parseDocstring)
    back to JSON (ie Python Dictionary)
    '''
    tag_dict = defaultdict(list)
    if doc is None : return tag_dict
    for child in doc.getchildren() :
        tag_dict[child.tag].append(child.text)

    return tag_dict




import os
import re

from lxml import etree
from ecldoc.Utils import write_to_file
from ecldoc.Utils import joinpath, dirname

###################################################################

from ecldoc.Constants import TEMPLATE_DIR
TXT_TEMPLATE_DIR = joinpath(TEMPLATE_DIR, 'txt')

###################################################################

import jinja2

txt_jinja_env = jinja2.Environment(
    loader = jinja2.FileSystemLoader(os.path.abspath('/'))
)

def indent_doc(s, level, width=4) :
    indention = (u' ' * (width-1) + u'| ') * level
    rv = (u'\n' + indention).join(s.splitlines())
    rv = indention + rv
    return rv

txt_jinja_env.filters['indent_doc'] = indent_doc

##################################################################

import textwrap

CPL = 100

def _break(text, CPL_E) :
    CPL_E = int(CPL_E)
    break_text = textwrap.wrap(text, CPL_E)
    if len(break_text) == 0 :
        break_text = ['']
    return break_text

##################################################################

from ecldoc.parseDoc import getTags
from ecldoc.Taglets import taglets
from .tagTXT import tag_renders

class ParseTXT(object) :
    '''
    Main class to generate TEXT Documentation for given ecl file
    from its XML Repr
    '''
    def __init__(self, generator, ecl_file) :
        self.xml_file = joinpath(generator.xml_root, ecl_file + '.xml')
        self.txt_file = joinpath(generator.txt_root, ecl_file + '.txt')
        self.template = generator.content_template
        self.options = generator.options

        os.makedirs(dirname(self.txt_file), exist_ok=True)

    def parse(self) :
        root = etree.parse(self.xml_file).getroot()
        src = root.find('Source')
        self.src = src
        self.doc = src.find('Documentation')

        for child in root.iter() :
            attribs = child.attrib
            ### Convert links from XML FOrmat to TXT Format
            if 'target' in attribs :
                attribs['target'] = re.sub(r'\$\$_ECLDOC-FORM_\$\$', 'txt', attribs['target'])
                attribs['target'] = re.sub(r'\.xml$', '.txt', attribs['target'])

        self.parseSource()

        render = self.template.render(src=src, defn_tree=self.defn_tree)
        write_to_file(self.txt_file, render)

    def docstring(self) :
        text = ''
        if self.doc is not None :
            content = self.doc.find('firstline')
            if content is not None :
                text = content.text
        return text

    def parseSource(self) :
        self.defn_tree = []
        for defn in self.src.findall('Definition') :
            self.parseDefinition(defn, self.defn_tree)

    def parseDefinition(self, defn, defn_tree) :
        headers = self.parseSign(defn)
        doc = self.parseDocs(defn)
        defn_dict = { 'headers' : headers, 'doc' : doc, 'defns' : [] }

        for childdefn in defn.findall('Definition') :
            self.parseDefinition(childdefn, defn_dict['defns'])

        defn_tree.append(defn_dict)

    def parseSign(self, defn) :
        defn_type = defn.attrib['type']
        sign = defn.find('Signature').text
        hlen = int(defn.find('Signature').attrib['hlen'])
        if defn.attrib['inherittype'] != 'local' :
            sign += ' ||| ' + defn.attrib['inherittype'].upper()

        heading = defn_type.upper() + ' : '
        spaces = len(heading)
        sign_break = _break(sign, CPL - spaces)
        type_break = [heading] + ([' '*(spaces + hlen)] * (len(sign_break) - 1))
        headers = [(a + b) for a,b in zip(type_break, sign_break)]
        return headers

    def parseDocs(self, defn) :
        renders = {}
        tags = getTags(defn.find('Documentation'))
        always = ['param', 'field', 'return', 'parent', 'content', 'firstline']
        common_tags = list(set(tags.keys()) | set(always))
        for tag in common_tags :
            if tag not in taglets or tag not in tag_renders :
                if 'generaltag' in tag_renders :
                    render = tag_renders['generaltag'](
                                taglets['generaltag'](doc=tags[tag], defn=defn, tagname=tag))
                    renders[tag] = render
                continue
            render = tag_renders[tag](taglets[tag](doc=tags[tag], defn=defn, tagname=tag))
            renders[tag] = render

        return renders

############################################################################

class GenTXT(object) :
    '''
    Generate TXT Documentation for all ecl files from XML Format
    '''
    def __init__(self, input_root, output_root, ecl_file_tree, options) :
        self.input_root = input_root
        self.output_root = output_root
        self.txt_root = joinpath(output_root, 'txt')
        self.xml_root = joinpath(output_root, 'xml')
        self.template_dir = TXT_TEMPLATE_DIR
        self.content_template = txt_jinja_env.get_template(joinpath(self.template_dir, 'content.tpl.txt'))
        self.toc_template = txt_jinja_env.get_template(joinpath(self.template_dir, 'toc.tpl.txt'))
        self.ecl_file_tree = ecl_file_tree
        self.options = options

    def gen(self, key, node, content_root) :
        '''
        Recursively parse source tree dictionary.
        If current_node is file : parse file using parseTXT
        Else If : current_node is directory : recurse and generate pkg.toc.txt for that dir
                                              (optionally generate bundle info if present)
        :param key: string | the name of current_node in path tree
        :param node: dict | the parent tree of current_node
        :param content_root: string | real path to current node in txt doc dir
        '''
        current_node = node[key]
        if type(current_node) != dict:
            if key == 'bundle.ecl' :
                return None
            parser = ParseTXT(self, current_node)
            parser.parse()
            file = { 'name' : key, 'target' : key + '.txt', 'type' : 'file', 'doc' : parser.docstring() }
            return file
        else :
            file = { 'name' : key,'target': joinpath(key, 'pkg.toc.txt'), 'type': 'dir', 'doc' : '' }

            bundle = None
            if 'bundle.ecl' in current_node :
                bundle_xml_path = joinpath(self.xml_root, dirname(current_node['bundle.ecl']), 'bundle.xml')
                bundle = etree.parse(bundle_xml_path).getroot()
                file['type'] = 'bundle'

            childfiles = []
            child_keys = sorted(current_node.keys(), key=str.lower)
            for chkey in child_keys :
                child_root = joinpath(content_root, chkey)
                child_dict = self.gen(chkey, current_node, child_root)
                if child_dict is not None : childfiles.append(child_dict)

            childfiles = sorted(childfiles, key=lambda x : x['type'])
            os.makedirs(content_root, exist_ok=True)
            render = self.toc_template.render(name=key, files=childfiles, bundle=bundle)
            render_path = joinpath(content_root, 'pkg.toc.txt')
            write_to_file(render_path, render)

            return file

    def run(self) :
        '''
        Main function called by ecldoc
        '''
        print("\nGenerating TEXT Documentation ... ")
        self.gen('root', self.ecl_file_tree, self.txt_root)



class tag(object) :
    def __init__(self, **kwargs) :
        self.docstrings = kwargs['doc']
        self.defn = kwargs['defn']
        self.tagname = kwargs['tagname'] if 'tagname' in kwargs else 'tag'

class tag_param(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.params = self.defn.find('Params')
        self.tuples = {'name' : 'Parameters', 'tuples' : []}
        self.create()

    def create(self) :
        broken_ = [c.split(' ', 1) for c in self.docstrings]
        broken_ = {c[0] : c[1] for c in broken_}
        params = self.params.findall('Param') if self.params is not None else []
        params = {p.attrib['name'].lower() : p for p in params}
        gen_params = {}
        for p in broken_ :
            param_name = p.lower().replace(':', '')
            actual = None
            if param_name in params :
                actual = params[param_name].find('Type')
                del params[param_name]

            gen_params[param_name] = [broken_[p], actual]
            self.tuples['tuples'].append((p, broken_[p], actual))

        for p in params :
            if p not in gen_params :
                self.tuples['tuples'].append((p, 'No Doc', params[p].find('Type')))

class tag_return(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.return_type = self.defn.find('Type')
        self.return_desc = self.docstrings[0] if len(self.docstrings) > 0 else ''

class tag_see(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.tuples = {'name' : 'See' , 'tuples' : []}
        for d in self.docstrings :
            self.tuples['tuples'].append((d, ))

class tag_field(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.fields = self.defn.findall('Field')
        self.tuples = {'name' : 'Fields', 'tuples' : []}
        self.create()

    def create(self) :
        broken_ = [c.split(' ', 1) for c in self.docstrings]
        broken_ = {c[0] : c[1] for c in broken_}
        fields = self.fields
        fields = {p.attrib['name'].lower() : p for p in fields}
        gen_fields = {}
        for p in broken_ :
            actual = None
            if p.lower() in fields :
                actual = fields[p.lower()].find('Type')
                del fields[p.lower()]

            gen_fields[p.lower()] = [broken_[p], actual]
            self.tuples['tuples'].append((p, broken_[p], actual))

        for p in fields :
            if p not in gen_fields :
                self.tuples['tuples'].append((p, 'No Doc', fields[p].find('Type')))

class tag_parent(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.parents = self.defn.find('Parents')
        self.tuples = {'name' : 'Parents', 'tuples' : []}
        self.create()

    def create(self) :
        parents = self.parents.findall('Parent') if self.parents is not None else []
        for p in parents :
            if 'ref' in p.attrib : self.tuples['tuples'].append((p.attrib['ref'], p.attrib['target']))

class tag_content(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.create()

    def create(self) :
        if len(self.docstrings) == 0 :
            self.text = 'No Documentation Found'
        else :
            self.text = " ".join(self.docstrings)

class tag_firstline(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.create()

    def create(self) :
        if len(self.docstrings) == 0 :
            self.text = "No Documentation Found"
        else :
            self.text = self.docstrings[0]

class tag_generaltag(tag) :
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)
        self.tuples = {'tuples' : [], 'name' : self.tagname}
        for d in self.docstrings :
            self.tuples['tuples'].append((d, ))

taglets= {  'param' : tag_param,
            'field' : tag_field,
            'see' : tag_see,
            'return' : tag_return,
            'parent' : tag_parent,
            'content' : tag_content,
            'firstline' : tag_firstline,
            'generaltag' : tag_generaltag
         }


#!/usr/bin/env python3

import os
import glob
import argparse
import configparser

from ecldoc.Utils import joinpath, relpath, realpath
from ecldoc.Utils import split

from ecldoc.genXML import GenXML
from ecldoc.Formats.generators import generators

def configParser(configfile) :
    '''
    Parse ecldoc options from configuration file (in INI Format) and append
    them to options dictionary

    :param configfile: path to configuration file
    '''
    options = {}

    cfgparser = configparser.ConfigParser()
    cfgparser.optionxform = str
    cfgparser.read(configfile)

    options['input_root'] = None
    if 'INPUT' in cfgparser.sections() :
        options['input_root'] = realpath(cfgparser['INPUT']['root'])
        options['nodoc'] = cfgparser['INPUT'].getboolean('hideNodoc', False)
        options['nointernal'] = cfgparser['INPUT'].getboolean('hideInternal', False)

        include = ['**/*.ecl']
        if 'include' in cfgparser['INPUT'] :
            include = split(cfgparser['INPUT']['include'], sep=',')
        options['include'] = include

        exclude = []
        if 'exclude' in cfgparser['INPUT'] :
            exclude = split(cfgparser['INPUT']['exclude'], sep=',')
        options['exclude'] = exclude

    options['output_root'] = None
    options['formats'] = ['html', 'text', 'pdf']
    if 'OUTPUT' in cfgparser.sections() :
        options['output_root'] = cfgparser['OUTPUT']['root']
        if 'format' in cfgparser['OUTPUT'] :
            options['formats'] = split(cfgparser['OUTPUT']['format'], sep=',', apply=lambda x : x.lower())

    options['eclcc'] = []
    if 'ECLCC' in cfgparser.sections() :
        for key in cfgparser['ECLCC'] :
            if cfgparser['ECLCC'].getboolean(key) :
                options['eclcc'].append(key)

    options['exdoc_paths'] = []
    if 'EXDOC' in cfgparser.sections() :
        paths = split(cfgparser['EXDOC']['paths'], sep=',')
        options['exdoc_paths'] = paths

    return options

def argsParserEcldoc(args) :
    '''
    Parse command line options and return as dictionary
    :param args: command line args returned by builtin ArgumentParser
    '''
    options = {}
    if args.config is not None :
        options = configParser(args.config)
    else :
        options['input_root'] = args.iroot
        options['output_root'] = args.oroot
        options['include'] = split(args.include, ',')
        options['exclude'] = split(args.exclude, ',')
        options['eclcc'] = split(args.eclcc, ';')
        options['nodoc'] = args.hideNoDoc
        options['nointernal'] = args.hideInternal
        options['exdoc_paths'] = split(args.exdocpaths, ',', apply=lambda x : realpath(x))
        options['formats'] = split(args.format, ',', apply=lambda x : x.lower())
    return options

def doMain() :
    '''
    Main Function (Entry point for ECLDOC)
    '''
    parser = argparse.ArgumentParser(description='Parser for ECLDOC')
    parser.add_argument('--config', help="Specify configuration file")
    parser.add_argument('-i', '--iroot', help="Specify input root")
    parser.add_argument('-o', '--oroot', help="Specify output root")
    parser.add_argument('-f', '--format', help="Specify output formats separated by , (eg html,text,pdf)", default='text')
    parser.add_argument('--include', help="Specify include paths seprated by ','", default='**/*.ecl')
    parser.add_argument('--exclude', help="Specify exclude paths seprataed by ','", default='')
    parser.add_argument('--eclcc', help="Specify eclcc Options separated by ';'", default='')
    parser.add_argument('--exdocpaths', help="Specify External Documentation paths separated by ','", default='')
    parser.add_argument('--hideNoDoc', help="Hide Definitions with No Documentation", action='store_true', default=False)
    parser.add_argument('--hideInternal', help="Hide Definitions with @internal tag", action='store_true', default=False)

    args = parser.parse_args()
    print(args)

    options = argsParserEcldoc(args)
    print(options)

    if options['input_root'] is None :
        print("Input Root Not Found")
        exit(1)
    if options['output_root'] is None :
        print("Output Root Not Found")
        exit(1)

    ### Apply include and exclude glob pattern to filter files in input root source tree
    ### ecl_files : Paths to filtered files stored in ecl_files list

    ecl_files = []
    input_root = realpath(options['input_root'])
    for pattern in options['include'] :
        pattern = joinpath(input_root, pattern)
        filenames = glob.glob(pattern, recursive=True)
        filenames = [realpath(f) for f in filenames]
        ecl_files += [relpath(f, input_root) for f in filenames]

    for pattern in options['exclude'] :
        pattern = joinpath(input_root, pattern)
        filenames = glob.glob(pattern, recursive=True)
        filenames = [realpath(f) for f in filenames]
        filenames = [relpath(f, input_root) for f in filenames]
        ecl_files = list(set(ecl_files) - set(filenames))

    ### Check if output root present and create it if not

    output_root = realpath(options['output_root'])
    if not os.path.exists(output_root) :
        os.makedirs(output_root, exist_ok=True)

    ### XML Intermediate format generator
    ### ecl_file_tree : Source tree is recreated from paths in ecl_files by XML Generator
    ###                 and filtered on basis of hideInternal and hideNoDoc

    xmlgenerator = GenXML(input_root, output_root, ecl_files, options)
    xmlgenerator.run()
    ecl_file_tree = xmlgenerator.ecl_file_tree

    ### Generator is run for each format that is present in format option by user

    for f in options['formats'] :
        if f in generators :
            generators[f](input_root, output_root, ecl_file_tree, options).run()

if __name__ == "__main__" :
    doMain()


import os, re, shutil

from lxml import etree
from ecldoc.Utils import getRoot, write_to_file
from ecldoc.Utils import joinpath, relpath, dirname

##################################################################

from ecldoc.Constants import TEMPLATE_DIR
HTML_TEMPLATE_DIR = joinpath(TEMPLATE_DIR, 'html')

##################################################################

import jinja2
html_jinja_env = jinja2.Environment(
    loader = jinja2.FileSystemLoader(os.path.abspath('/'))
)

#################################################################

from ecldoc.parseDoc import getTags
from ecldoc.Taglets import taglets
from .tagHTML import tag_renders

class ParseHTML(object) :
    '''
    Main class to generate HTML Documentation for given ecl file
    from its XML Repr
    '''
    def __init__(self, generator, ecl_file) :
        self.output_root = generator.output_root
        self.xml_file = joinpath(generator.xml_root, (ecl_file + '.xml'))
        self.html_file = joinpath(generator.html_root, (ecl_file + '.html'))
        self.template = generator.content_template

        ### Parent node of given file in path tree
        self.parent = getRoot(generator.ecl_file_tree, ecl_file)
        self.options = generator.options

        os.makedirs(dirname(self.html_file), exist_ok=True)

    def parse(self) :
        tree = etree.parse(self.xml_file)
        root = tree.getroot()
        src = root.find('Source')
        self.src = src
        self.doc = src.find('Documentation')

        for child in root.iter() :
            attribs = child.attrib
            ### Convert links from XML FOrmat to HTML Format
            if 'target' in attribs :
                attribs['target'] = re.sub(r'\$\$_ECLDOC-FORM_\$\$', 'html', attribs['target'])
                attribs['target'] = re.sub(r'\.xml$', '.html', attribs['target'])

            ### Fullname act as HTML Id for Definitions
            ### '.' is invalid char in HTML Id so replaced by '-'
            if 'fullname' in attribs :
                attribs['origfn'] = attribs['fullname']
                attribs['fullname'] = re.sub(r'\.', '-', attribs['fullname'])

        self.parseSource()

        files = []
        keys = sorted(self.parent.keys(), key=str.lower)
        for key in keys :
            if type(self.parent[key]) != dict :
                file = {'name': key,
                        'target': key + '.html',
                        'type': 'file'}
                files.append(file)
            else :
                file = {'name': key,
                        'target': joinpath(key, 'pkg.toc.html'),
                        'type': 'dir'}
                files.append(file)

        files = sorted(files, key=lambda x : x['type'])

        parent = 'pkg.toc.html'
        output_relpath = relpath(self.output_root, dirname(self.html_file))
        render = self.template.render(src=src,
                                      files=files,
                                      parent=parent,
                                      output_root=output_relpath,
                                      defn_tree=self.defn_tree)
        write_to_file(self.html_file, render)

    def docstring(self) :
        text = ''
        if self.doc is not None :
            content = self.doc.find('firstline')
            if content is not None :
                text = content.text
        return text

    def parseSource(self) :
        self.defn_tree = []
        for defn in self.src.findall('Definition') :
            self.parseDefinition(defn, self.defn_tree)

    def parseDefinition(self, defn, defn_tree) :
        doc = self.parseDocs(defn)
        defn_dict = { 'tag' : defn, 'doc' : doc, 'defns' : [] }

        for childdefn in defn.findall('Definition') :
            self.parseDefinition(childdefn, defn_dict['defns'])

        defn_tree.append(defn_dict)

    def parseDocs(self, defn) :
        '''
        Convert Doctags into HTML Format
        '''
        renders = {}
        tags = getTags(defn.find('Documentation'))
        always = ['param', 'field', 'return', 'parent', 'content', 'firstline']
        common_tags = list(set(tags.keys()) | set(always))
        for tag in common_tags :
            if tag not in taglets or tag not in tag_renders :
                if 'generaltag' in tag_renders :
                    render = tag_renders['generaltag'](
                                taglets['generaltag'](doc=tags[tag], defn=defn, tagname=tag))
                else :
                    render = tags[tag]
                renders[tag] = render
                continue
            render = tag_renders[tag](taglets[tag](doc=tags[tag], defn=defn, tagname=tag))
            renders[tag] = render

        return renders


class GenHTML(object) :
    '''
    Generate HTML Documentation for all ecl files from XML Format
    '''
    def __init__(self, input_root, output_root, ecl_file_tree, options) :
        self.input_root = input_root
        self.output_root = output_root
        self.html_root = joinpath(output_root, 'html')
        self.xml_root = joinpath(output_root, 'xml')

        self.template_dir = HTML_TEMPLATE_DIR
        self.content_template = html_jinja_env.get_template(joinpath(self.template_dir, 'content.tpl.html'))
        self.toc_template = html_jinja_env.get_template(joinpath(self.template_dir, 'toc.tpl.html'))

        self.ecl_file_tree = ecl_file_tree
        self.options = options

    def gen(self, key, node, content_root) :
        '''
        Recursively parse source tree dictionary.
        If current_node is file : parse file using parseHTML
        Else If : current_node is directory : recurse and generate pkg.toc.html for that dir
                                              (optionally generate bundle info if present)
        :param key: string | the name of current_node in path tree
        :param node: dict | the parent tree of current_node
        :param content_root: string | real path to current node in html doc dir
        '''
        current_node = node[key]
        if type(current_node) != dict:
            if key == 'bundle.ecl' :
                return None
            parser = ParseHTML(self, node[key])
            parser.parse()
            file = {'name' : key,
                    'target' : key + '.html',
                    'type' : 'file',
                    'doc' : parser.docstring() }
            return file
        else :
            file = {'name' : key,
                    'target': joinpath(key, 'pkg.toc.html'),
                    'type': 'dir',
                    'doc' : '' }

            bundle = None
            if 'bundle.ecl' in current_node :
                bundle_xml_path = joinpath(self.xml_root, dirname(current_node['bundle.ecl']), 'bundle.xml')
                bundle = etree.parse(bundle_xml_path).getroot()
                license = bundle.find('License')
                license.text = '<a href="' + license.text + '">' + license.text + '</a>'
                file['type'] = 'bundle'

            childfiles = []
            keys = sorted(current_node.keys(), key=str.lower)
            for chkey in keys :
                child_root = joinpath(content_root, chkey)
                child_dict = self.gen(chkey, current_node, child_root)
                if child_dict is not None : childfiles.append(child_dict)

            childfiles = sorted(childfiles, key=lambda x : x['type'])

            root_relpath = relpath(self.output_root, content_root)
            parent_relpath = ''
            if content_root != self.html_root :
                parent_relpath = relpath(dirname(content_root), content_root)

            render = self.toc_template.render(name=key,
                                              files=childfiles,
                                              parent=joinpath(parent_relpath, 'pkg.toc.html'),
                                              output_root=root_relpath,
                                              bundle=bundle)
            os.makedirs(content_root, exist_ok=True)
            render_path = joinpath(content_root, 'pkg.toc.html')
            write_to_file(render_path, render)

            return file

    def run(self) :
        '''
        Main function called by ecldoc
        '''
        print("\nGenerating HTML Documentation ... ")
        self.gen('root', self.ecl_file_tree, self.html_root)

        if os.path.exists(joinpath(self.output_root, 'css')) :
            shutil.rmtree(joinpath(self.output_root, 'css'))
        shutil.copytree(joinpath(self.template_dir, 'css'), joinpath(self.output_root, 'css'))


import os
import re
import json
import subprocess
from copy import deepcopy

from lxml import etree
from lxml.builder import E

from .Utils import genPathTree
from .Utils import joinpath, relpath, dirname, realpath
from .Utils import read_file

from .parseDoc import parseDocstring, cleansign, breaksign


class ParseXML(object):
    '''
    ParseXML is main class to generate XML Documentation for a single ecl file
    '''
    def __init__(self, generator, ecl_file):
        '''
        :param generator: GenXML object creating this object
        :param ecl_file: path to ecl_file to parse
        '''
        self.input_root = generator.input_root
        self.output_root = generator.output_root
        self.ecl_file = ecl_file

        ### complete path to ecl file
        self.input_file = joinpath(self.input_root, ecl_file)

        ### Path to file where output from eclcc -M will be stored (called xmlOriginal)
        self.xml_orig_root = generator.xml_orig_root
        self.xml_orig_file = joinpath(self.xml_orig_root, ecl_file + '.xml')

        ### Path to file where processed xml from xmlOriginal will be stored
        self.xml_root = generator.xml_root
        self.xml_file = joinpath(self.xml_root, ecl_file + '.xml')
        self.xml_dir = dirname(self.xml_file)

        self.generator = generator
        self.options = generator.options

        ### Check if any definitions remaining in xml tree after processing
        self.internal = False

        #local
        self.depends = {}
        self.src = None

    def parse(self):
        '''
        Main function called by generator on this object
        Generates both XMLOriginal and processed XML Doc
        '''

        ### Run eclcc -M on input_file to generate xmlOriginal doc
        os.makedirs(dirname(self.xml_orig_file), exist_ok=True)
        eclcc_options = ' '.join(self.options['eclcc'])
        p = subprocess.call(['eclcc -M ' + eclcc_options +
                             ' -o ' + self.xml_orig_file + ' ' +
                             self.input_file], shell=True, cwd=self.input_root)
        print("ECLCC ||| File : ", self.ecl_file, " ||| Output Code : ", p)

        if p > 0 :
            print("ECLCC Error : Removing file from documentation source tree")
            self.internal = True
            return

        tree = etree.parse(self.xml_orig_file)
        root = tree.getroot()

        ### Main loop of this function
        ### Loops over all Source Tags in xmlOriginal

        for src in root.iter('Source'):
            attribs = src.attrib
            srcpath = realpath(attribs['sourcePath'])

            ### Check each Source Tag in xmlOriginal to see if it match input_file
            ### If yes, keep it, else replace that Source Tag with Depend Tag

            if os.path.exists(srcpath) and srcpath.startswith(self.input_root):
                attribs['sourcePath'] = srcpath
                src_relpath = relpath(srcpath, self.input_root)
                tgtpath = joinpath(self.xml_root, (src_relpath + '.xml'))
                tgtpath = realpath(tgtpath)
                tgtpath = relpath(tgtpath, self.xml_dir)
                attribs['target'] = tgtpath
                if src_relpath == self.ecl_file:
                    continue

            root.remove(src)
            depend = etree.Element('Depends', sourcePath=attribs['sourcePath'])

            if 'name' in attribs:
                depend.set('name', attribs['name'])

                ### Create a entry in depend dictionary for current Source
                ### Used to map Imports and other targets to correct file
                self.depends[tuple(attribs['name'].lower().split('.'))] = depend

            ### Set Target Path to processed XML for given Dependency
            ### Match its source path to current source tree and external documentation (if any)
            if 'target' in attribs:
                depend.set('target', attribs['target'])
            else:
                if 'name' in attribs:
                    matched = self.generator.matchExdoc(attribs['name'], source=attribs['sourcePath'])
                    if matched is not None:
                        depend.set('target', matched + '.xml')
                    else:
                        depend.set('target', attribs['sourcePath'])
                else:
                    depend.set('target', attribs['sourcePath'])

            root.insert(-1, depend)

        ### Only Source Tag corresponding to input_file is left and is processed
        self.src = root.find('Source')
        self.parseSource()

        ### If no definitions left, set internal to True
        if len(self.src.findall('Definition')) == 0:
            self.internal = True

        ### write processed XML to xml_file if not internal
        if self.internal is False:
            os.makedirs(dirname(self.xml_file), exist_ok=True)
            tree.write(self.xml_file, xml_declaration=True, encoding='utf-8')

    def parseSource(self):
        attribs = self.src.attrib
        srcpath = realpath(attribs['sourcePath'])
        print('parseSource srcpath = ', srcpath)
        self.text = read_file(srcpath);

        ### Append to depend dictionary entry correspong to given file
        ###        for self referential targets
        if 'name' in attribs:
            self.depends[tuple(attribs['name'].lower().split('.'))] = self.src

        for imp in self.src.iter('Import'):
            self.parseImport(imp)

        for doc in self.src.iter('Documentation'):
            self.parseDocumentation(doc)

        for defn in self.src.findall('Definition'):
            if self.checkDefinition(defn):
                self.src.remove(defn)
            else:
                self.parseDefinition(defn)

        ### Documentation for first Definition in Source become Documentation
        ###     for the source. Useful in displaying documentation in TOCs
        maindefn = self.src.find("Definition")
        if maindefn is not None and maindefn.find('Documentation') is not None:
            docstring = deepcopy(maindefn.find('Documentation'))
            self.src.append(docstring)
        else:
            self.src.append(E('Documentation', E('content', ' ')))

    def parseDefinition(self, defn):
        print('parseDefinition defn = ', defn)
        attribs = defn.attrib
        self.generateSignature(defn)

        if 'fullname' in attribs:
            fullname = attribs['fullname']
            best_depend = self.matchPath(fullname)
            if best_depend is not None and best_depend != self.src:
                attribs['target'] = best_depend.attrib['target']
        elif 'name' in attribs:
            attribs['fullname'] = 'ecldoc-' + attribs['name']

        for childdefn in defn.findall('Definition'):
            ### Do not parse children of inherited attributes, in addition to normal checks
            if attribs['inherittype'] == 'inherited' or self.checkDefinition(childdefn):
                defn.remove(childdefn)
            else:
                self.parseDefinition(childdefn)

        parents = defn.find('Parents')
        if parents is not None:
            self.parseParents(parents)

        ### Remove unnecessary attributes
        attribs.pop('body', None)
        attribs.pop('start', None)
        attribs.pop('end', None)

    def generateSignature(self, defn):
        '''
        Generate attribute signature by reading the ecl source file where
        given Definition is present.
        Signature is text between EXPORT and :=
        '''
        attribs = defn.attrib
        name = attribs['name']
        print('generateSignature name = ', name)
        is_scope = 'type' in attribs and attribs['type'] in ['module', 'interface']
        has_params = defn.find('Params') is not None

        ecl_text = self.text
        gen_sign = False

        if 'fullname' in attribs:
            gen_sign = True
            ### MORE : if definition is scope without params, do not read file for it.
            ###        Problem in fullname resolution
            if is_scope and (not has_params) :
                gen_sign = False
            else :
                ### If definition signature in external file (eg inherited attributes)
                ### based on fullname, Find that file
                best_depend = self.matchPath(attribs['fullname'])
                if best_depend is not None and best_depend != self.src:
                    print( 'generateSignature fullname = ', best_depend.attrib['sourcePath'])
                    ecl_text = read_file(best_depend.attrib['sourcePath']);

        sign = etree.Element('Signature')
        sign.text = name
        if gen_sign and ('start' in attribs) and ('body' in attribs) :
            sign.text = ecl_text[int(attribs['start'])-2:int(attribs['body'])]
            sign.text = cleansign(sign.text)

        ### break sign text into return type, name and parameter segments (Heuristic only)
        ### Indent len = strlen(ret) + strlen(name)
        ret, param, indent_len = breaksign(name, sign.text)

        sign.attrib['name'] = name
        sign.attrib['ret'] = ret
        sign.attrib['param'] = param
        sign.attrib['hlen'] = str(indent_len)
        defn.insert(-1, sign)

    def parseDocumentation(self, doc):
        content = doc.find('content')
        #print('parseDocumentation doc = ', etree.tostring(doc))
        #print('parseDocumentation content = ', etree.tostring(content))
        elements = parseDocstring(content.text)
        doc.remove(content)
        ### Append all parsed doctags from docstring as XML Tags
        for tag in elements:
          for desc in elements[tag]:
            doc.append(desc)

    def parseImport(self, imp):
        attribs = imp.attrib
        ### Match Reference of Import to correct File/Directory Path for Linking
        if 'ref' in attribs:
            attribs['target'] = self.matchReference(attribs['ref'])

        attribs.pop('body', None)
        attribs.pop('start', None)
        attribs.pop('end', None)
        attribs.pop('fullname', None)
        attribs.pop('line', None)
        attribs.pop('inherittype', None)
        if imp.find('Documentation') is not None :
            imp.remove(imp.find('Documentation'))

    def parseParents(self, parents):
        for parent in parents.findall('Parent'):
            attribs = parent.attrib
            ### Match Reference (Fullname) of parent to correct File/Directory Path for Linking
            if 'ref' in attribs:
                attribs['target'] = self.matchReference(attribs['ref'])

    def matchReference(self, refpath):
        '''
        Match reference/fullname to correct file/dir path
        '''
        target = ''
        ### First search among Dependency Files
        best_depend = self.matchPath(refpath)
        if best_depend is not None:
            target = best_depend.attrib['target']
        else:
            ### Then search in External Documentation (assumption : only directory paths left)
            matched = self.generator.matchExdoc(refpath)
            if matched is not None:
                target = matched + '.xml'
            else:
                ### Simply Convert reference to path (assumption : only directory paths left)
                refpath = refpath.split('.') + ['pkg.toc.xml']
                target = joinpath(self.xml_root, *refpath)
                target = relpath(target, self.xml_dir)
        return target

    def matchPath(self, path):
        '''
        Find best prefix match for given fullname/reference in Depend Dictionary
        '''
        path = tuple(path.lower().split('.'))
        current_prefix = ()
        for depend_path in self.depends:
            prefix = os.path.commonprefix([path, depend_path])
            if len(prefix) > len(current_prefix):
                current_prefix = prefix

        if len(current_prefix) > 0 and current_prefix in self.depends:
            return self.depends[current_prefix]

        return None

    def checkDefinition(self, defn):
        '''
        Check if definition should be expanded or removed
        True : Definition should be removed
        '''
        test_1 = 'exported' not in defn.attrib
        test_2 = self.options['nodoc'] and (defn.find('Documentation') is None)
        test_3 = (self.options['nointernal'] and
                  (defn.find('Documentation') is not None) and
                  (defn.find('Documentation').find('internal') is not None))
        return test_1 or test_2 or test_3

###########################################################################################

def parseBundle(generator, ecl_file):
    input_file = joinpath(generator.input_root, ecl_file)
    dirpath = dirname(input_file)

    dirpath_xml_orig = dirname(joinpath(generator.xml_orig_root, ecl_file))
    dirpath_xml = dirname(joinpath(generator.xml_root, ecl_file))

    bundle_orig_path = joinpath(dirpath_xml_orig, 'bundle.orig.out')
    bundle_xml_path = joinpath(dirpath_xml, 'bundle.xml')

    os.makedirs(dirpath_xml_orig, exist_ok=True)
    p = subprocess.call(['ecl-bundle info ' + dirpath + ' > ' + bundle_orig_path], shell=True)
    print("ECL-BUNDLE ||| ", "Bundle File : ", dirpath, "Output Code : ", p)

    data = read_file(bundle_orig_path).split('\n')
    data = [x.split(':', 1) for x in data]
    data = [(x[0].strip(), x[1].strip()) for x in data if len(x) == 2]
    root = etree.Element("Bundle")
    for k in data:
        node = etree.Element(k[0])
        node.text = k[1]
        root.append(node)

    os.makedirs(dirname(bundle_xml_path), exist_ok=True)
    etree.ElementTree(root).write(bundle_xml_path, pretty_print=True, xml_declaration=True, encoding='utf-8')

###########################################################################################

class GenXML(object):
    '''
    Generate XML Doc all files in ecl_files.
    Two types of XML Docs : xmlOriginal (eclcc output) and XML (after processing)
    '''
    def __init__(self, input_root, output_root, ecl_files, options):
        self.input_root = input_root
        self.output_root = output_root
        ### Recreate Source Tree as dictionary from ecl file paths
        self.ecl_file_tree = genPathTree(ecl_files)
        self.options = options

        ### Folder where xmlOriginal files are stored
        self.xml_orig_root = joinpath(output_root, 'xmlOriginal')
        os.makedirs(self.xml_orig_root, exist_ok=True)

        ### Folder where processed XML files are stored
        self.xml_root = joinpath(output_root, 'xml')
        os.makedirs(self.xml_root, exist_ok=True)

    def gen(self, key, node, xml_root, content_root):
        '''
        Recursively parse source tree dictionary.
        If current_node is file : parse file using parseXML or parseBundle
        Else If : current_node is directory : recurse and generate pkg.toc.xml for that dir
        :param key: string | the name of current_node in path tree
        :param node: dict | the parent of current_node (parent is passed so that current_node
        					can be deleted if it is internal file)
        :param content_root: string | real path to current node
        :param xml_root: lxml.Element | current node in xml repr of path tree for TOCs
        '''
        current_node = node[key]
        if type(current_node) != dict:
            if key == 'bundle.ecl':
                parseBundle(self, current_node)
                return
            ecl_file = current_node
            parser = ParseXML(self, ecl_file)
            parser.parse()
            if parser.internal:
                del node[key]
            else:
                file = etree.Element('file')
                file.attrib['name'] = key + '.xml'
                file.text = current_node
                xml_root.append(file)
        else:
            folder = etree.Element('folder')
            xml_root.append(folder)
            folder.attrib['name'] = key

            child_keys = list(current_node.keys())
            for chkey in child_keys:
                child_root = joinpath(content_root, chkey)
                self.gen(chkey, current_node, folder, child_root)

            os.makedirs(content_root, exist_ok=True)
            toc_file = joinpath(content_root, 'pkg.toc.xml')
            etree.ElementTree(folder).write(toc_file, pretty_print=True, xml_declaration=True, encoding='utf-8')
            self.genJsonTree(current_node, content_root)

    def run(self):
        '''
        Main function called by ecldoc
        '''
        print("\nGenerating XML Documentation ... ")
        self.processExternalDoc()
        root = etree.Element('Root')
        self.gen('root', self.ecl_file_tree, root, self.xml_root)

    #######################################################################
    # Generate tree for files and subfolders in each folder for given run #
    # Currently used by External Dodc matching system .                   #
    #######################################################################

    def genJsonTree(self, tree, output_path='', dump=True):
        json_output = {}
        new_tree = deepcopy(tree)
        self.processJsonTree(new_tree)
        json_output['tree'] = new_tree
        json_output['output_root'] = self.output_root
        json_output['input_root'] = self.input_root
        if output_path != '' :
        	rel_output_path = relpath(output_path, self.xml_root)
        	json_output['include_path'] = os.path.normpath(joinpath(self.input_root, rel_output_path))
        else :
        	json_output['include_path'] = self.input_root
        if dump:
            json.dump(json_output, open(joinpath(output_path, 'tree.json'), 'w'), sort_keys=True, indent=4)

        return json_output

    def processJsonTree(self, tree):
        keys = list(tree.keys())
        for key in keys:
            if type(tree[key]) == dict:
                self.processJsonTree(tree[key])
            node = tree[key]
            del tree[key]
            tree[re.sub(r'\.ecl$', '', key.lower())] = {'key': key, 'tree': node}

    ###############################################################
    # Process External Docs and provide match function for parser #
    ###############################################################

    def processExternalDoc(self):
        self.exdocs = []
        currdoc = self.genJsonTree(self.ecl_file_tree['root'], dump=False)
        self.exdocs.append(currdoc)
        for path in self.options['exdoc_paths']:
            json_file = joinpath(path, 'tree.json')
            assert os.path.isfile(json_file), ("Exdoc file does not exists : " + json_file)
            external_doc = json.load(open(json_file))
            self.exdocs.append(external_doc)

    def matchExdoc(self, fullname, source=''):
        fullname = tuple(fullname.lower().split('.'))
        matched = None
        for exdoc in self.exdocs:
            val = self.findInTree(fullname, exdoc['tree'], '')
            if val is not None:
                if source == '' or realpath(source).startswith(exdoc['input_root']):
                    matched = joinpath(exdoc['output_root'], '$$_ECLDOC-FORM_$$', val)
                    break
        return matched

    def findInTree(self, fullname, tree, child_root):
        if type(tree) == dict:
            if len(fullname) == 0: return joinpath(child_root, 'pkg.toc')
            if fullname[0] in tree:
                return self.findInTree(fullname[1:],
                                       tree[fullname[0]]['tree'],
                                       joinpath(child_root, tree[fullname[0]]['key']))
            return None

        return tree



from .genTEX import latex_jinja_env, TEX_TEMPLATE_DIR, escape_tex
from ecldoc.Utils import call_macro_by_name, joinpath
from ecldoc.parseDoc import construct_type
import lxml.html as H

latex_jinja_env.filters['macro'] = call_macro_by_name
tag_template = latex_jinja_env.get_template(joinpath(TEX_TEMPLATE_DIR, 'taglets.tpl.tex'))

###############################################################

def convertToLatex(html_text) :
    '''
    Convert HTML String to Markdown Format
    '''
    root = H.fragment_fromstring(html_text, create_parent='div')
    text = parseHTMltoLatex(root)
    return text

def parseHTMltoLatex(element) :
    '''
    Convert Single HTML element to Markdown Text (recursive)
    '''
    text = ''
    if element.text :
        text += escape_tex(element.text)
    for e in element.iterchildren() :
        text += parseHTMltoLatex(e)
        if e.tail :
            text += escape_tex(e.tail)

    if element.tag in ['p'] :
        text = '\n\\par\n' + text + '\n\n'

    if element.tag == 'br' :
        text = '\\\\\n'

    if element.tag == 'br' and element.getparent().tag == 'pre' :
        text = '\n'

    if element.tag == 'code' :
        text = '```' + text  + '```'

    if element.tag == 'li' :
        text = '\\item ' + text + '\n'

    if element.tag == 'ul' :
        text = '\n\\begin{itemize}\n' + text + '\\end{itemize}\n\n'

    if element.tag == 'ol' :
        text = '\n\\begin{enumerate}\n' + text + '\\end{enumerate}\n\n'

    if element.tag == 'a' :
        text = '\n\\url{' + element.attrib['href'] + '}'

    if element.tag == 'pre' :
        text = '\n\\begin{verbatim}\n' + text + '\\end{verbatim}\n\n'

    return text

###############################################################

def render_param(tag_param) :
    if len(tag_param.tuples['tuples']) == 0 : return ''
    tex_tuples = []
    for t in tag_param.tuples['tuples'] :
        tex_tuples.append((t[0], construct_type(t[2]), t[1]))

    render = tag_template.render(render_name='threetag', args=['Parameter', tex_tuples])
    return render

def render_field(tag_field) :
    if len(tag_field.tuples['tuples']) == 0 : return ''
    html_tuples = []
    for t in tag_field.tuples['tuples'] :
        html_tuples.append((t[0], construct_type(t[2]), t[1]))

    return tag_template.render(render_name='threetag', args=['Field', html_tuples])

def render_return(tag_return) :
    if len(tag_return.docstrings) == 0 and tag_return.return_type is None : return ''
    return_tuple = [(construct_type(tag_return.return_type), tag_return.return_desc)]
    return tag_template.render(render_name='twotag', args=['Return', return_tuple])

def render_see(tag_see) :
    render = tag_template.render(render_name='onetag', args=['See', tag_see.tuples['tuples']])
    return render

def render_parent(tag_parent) :
    if len(tag_parent.tuples['tuples']) == 0 : return ''
    render = tag_template.render(render_name='linktag', args=['Parent', tag_parent.tuples['tuples']])
    return render

def render_content(tag_content) :
    text_latex = convertToLatex(tag_content.text).split('\n')
    render = tag_template.render(render_name='contenttag', args=[text_latex])
    return render

def render_firstline(tag_firstline) :
    text_latex = convertToLatex(tag_firstline.text)
    return text_latex

def render_inherit(tag_inherit) :
    if tag_inherit == 'local' : return ''
    render = tag_template.render(render_name='onetag', args=[tag_inherit, [('', )]])
    return render

def render_generaltag(tag_generaltag) :
    render = tag_template.render(render_name='onetag',
                                args=[tag_generaltag.tuples['name'], tag_generaltag.tuples['tuples']])
    return render

tag_renders = { 'param' : render_param,
                'field' : render_field,
                'return' : render_return,
                'see' : render_see,
                'parent' : render_parent,
                'content' : render_content,
                'firstline' : render_firstline,
                'inherit' : render_inherit,
                'generaltag' : render_generaltag
              }



from ecldoc.parseDoc import construct_type
from .genTXT import CPL, _break
import lxml.html as H

##########################################################

def convertToMarkdown(html_text) :
    '''
    Convert HTML String to Markdown Format
    '''
    root = H.fragment_fromstring(html_text, create_parent='div')
    text = parseHTMltoMK(root)
    return text

def parseHTMltoMK(element) :
    '''
    Convert Single HTML element to Markdown Text (recursive)
    '''
    text = ''
    if element.text :
        text += element.text
    for e in element.iterchildren() :
        text += parseHTMltoMK(e)
        if e.tail :
            text += e.tail

    if element.tag in ['p', 'pre', 'ul', 'ol', 'table'] :
        text = '\n' + text + '\n'

    if element.tag == 'br' :
        text = '\n'

    if element.tag == 'code' :
        text = '```' + text  + '```'

    if element.tag == 'li' and element.getparent().tag == 'ul':
        text = '+ ' + text + '\n'

    if element.tag == 'li' and element.getparent().tag == 'ol':
        text = '# ' + text + '\n'

    if element.tag == 'td' :
        text = text + ' | '

    if element.tag == 'tr' :
        text = '| ' + text + '\n'

    if element.tag == 'hr' :
        text = '\n************\n'

    if element.tag == 'a' :
        text = text + ' <' + element.attrib['href'] + '>'

    return text

##########################################################

def parseTag(text, heading) :
    heading = heading
    text_break = _break(text, CPL * 0.5)
    spaces = [heading] + ([' ' * len(heading)] * (len(text_break)-1))
    text_list = [(a + b) for a, b in zip(spaces, text_break)]
    return text_list

def render_param(tag_param) :
    lines = []
    for t in tag_param.tuples['tuples'] :
        lines.append(parseTag(construct_type(t[2]) + ' --- ' + t[1], 'Parameter : ' + t[0] + ' '))
    return lines

def render_field(tag_field) :
    lines = []
    for t in tag_field.tuples['tuples'] :
        lines.append(parseTag(construct_type(t[2]) + ' --- ' + t[1], 'Field : ' + t[0] + ' '))
    return lines

def render_return(tag_return) :
    lines = []
    if len(tag_return.docstrings) == 0 and tag_return.return_type is None : return lines
    lines.append(parseTag(construct_type(tag_return.return_type) + ' ' + tag_return.return_desc, 'Return : '))
    return lines

def render_see(tag_see) :
    lines = []
    for t in tag_see.tuples['tuples'] :
        lines.append(parseTag(t[0], 'See : '))
    return lines

def render_parent(tag_parent) :
    lines = []
    for t in tag_parent.tuples['tuples'] :
        lines.append(parseTag(t[0] + ' <' + t[1] + '>', 'Parent : '))
    return lines

def render_content(tag_content) :
    mktext = convertToMarkdown(tag_content.text).split('\n')
    lines = []
    for t in mktext :
        lines += _break(t, CPL*0.75)
    return lines

def render_firstline(tag_firstline) :
    lines = []
    lines.append([tag_firstline.text])
    return lines

def render_generaltag(tag_generaltag) :
    lines = []
    for t in tag_generaltag.tuples['tuples'] :
        lines.append(parseTag(t[0], tag_generaltag.tuples['name'] + ' : '))
    return lines

tag_renders = { 'param' : render_param,
                'field' : render_field,
                'return' : render_return,
                'see' : render_see,
                'parent' : render_parent,
                'content' : render_content,
                'firstline' : render_firstline,
                'generaltag' : render_generaltag
              }



from .genHTML import html_jinja_env, HTML_TEMPLATE_DIR
from ecldoc.Utils import call_macro_by_name, joinpath
from ecldoc.parseDoc import construct_type

html_jinja_env.filters['macro'] = call_macro_by_name
tag_template = html_jinja_env.get_template(joinpath(HTML_TEMPLATE_DIR, 'taglets.tpl.html'))

def render_param(tag_param) :
    if len(tag_param.tuples['tuples']) == 0 : return ''
    html_tuples = []
    for t in tag_param.tuples['tuples'] :
        html_tuples.append((t[0], construct_type(t[2]), t[1]))

    render = tag_template.render(render_name='threetag', args=['Parameters', html_tuples])
    return render

def render_field(tag_field) :
    if len(tag_field.tuples['tuples']) == 0 : return ''
    html_tuples = []
    for t in tag_field.tuples['tuples'] :
        html_tuples.append((t[0], construct_type(t[2]), t[1]))

    return tag_template.render(render_name='threetag', args=['Fields', html_tuples])

def render_return(tag_return) :
    if len(tag_return.docstrings) == 0 and tag_return.return_type is None : return ''
    return_tuple = [(construct_type(tag_return.return_type), tag_return.return_desc)]
    return tag_template.render(render_name='twotag', args=['Returns', return_tuple])

def render_see(tag_see) :
    render = tag_template.render(render_name='onetag', args=['See', tag_see.tuples['tuples']])
    return render

def render_parent(tag_parent) :
    if len(tag_parent.tuples['tuples']) == 0 : return ''
    render = tag_template.render(render_name='parenttag', args=['Parents', tag_parent.tuples['tuples']])
    return render

def render_content(tag_content) :
    render = tag_template.render(render_name='contenttag', args=[tag_content.text])
    return render

def render_firstline(tag_firstline) :
    return tag_firstline.text

def render_generaltag(tag_generaltag) :
    render = tag_template.render(render_name='onetag', args=[tag_generaltag.tuples['name'], tag_generaltag.tuples['tuples']])
    return render

tag_renders = { 'param' : render_param,
                'field' : render_field,
                'return' : render_return,
                'see' : render_see,
                'parent' : render_parent,
                'content' : render_content,
                'firstline' : render_firstline,
                'generaltag' : render_generaltag
              }






from setuptools import setup, find_packages

import os


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


extra_files = package_files('ecldoc/Templates')
print(extra_files)
setup(
    name="ecldoc",
    version="2.0.0",
    packages=find_packages(),
    install_requires=[
        'Jinja2==2.9.6',
        'lxml==4.9.0',
        'importlib-metadata==4.8.3',
        'markdown==3.3.7',
        'typing-extensions==4.1.1',
        'zipp==3.6.0',
        'webencodings==0.5.1',
        'zopfli==0.1.9',
        'CairoSVG==2.5.2',
        'Pillow==8.0.0',
        'Pyphen==0.11.0',
        'cffi==1.15.0',
        'cssselect2==0.4.1',
        'defusedxml==0.7.1',
        'html5lib==1.1',
        'pycparser==2.21',
        'setuptools==59.6.0',
        'six==1.16.0',
        'tinycss2==1.1.1',
        'weasyprint==52.5',
        'webencodings==0.5.1',
    ],
    package_data={'': extra_files},
    scripts=['bin/ecldoc'],
    url="https://github.com/lilyclemson/ECLDocGenerator",
)




import os

########################################################################

def split(arg, sep, apply=lambda x : x) :
    return [apply(x.strip()) for x in arg.split(sep) if len(x) != 0]

########################################################################

from jinja2 import contextfilter

@contextfilter
def call_macro_by_name(context, macro_name, *args, **kwargs):
    return context.vars[macro_name](*args, **kwargs)

########################################################################

def genPathTree(ecl_files, ext='') :
    '''
    Convert filepaths recovered from glob pattern into a tree structure
    '''
    path_tree = { "root" : {} }
    for file in ecl_files :
        path = os.path.normpath(file).split(os.sep)
        parent = path_tree["root"]
        for node in path[:-1] :
            if node not in parent :
                parent[node] = {}
            parent = parent[node]

        if path[-1].lower() == 'bundle.ecl' :
            path[-1] = 'bundle.ecl'
        parent[path[-1]] = file + ext

    return path_tree

def getRoot(path_tree, ecl_file) :
    '''
    walk the filepath tree to get parent of ecl_file
    '''
    path = os.path.normpath(ecl_file).split(os.sep)[:-1]
    parent_path = ''
    parent = path_tree['root']
    for node in path :
        parent = parent[node]
        parent_path = os.path.join(parent_path, node)

    return parent

########################################################################

def write_to_file(filename, text) :
    fp = open(filename, 'w')
    fp.write(text)
    fp.close()

def read_file(filename) :
    # Note the newline = ''.  This is required in order to prevent
    # <cr><nl> from being changed to <nl> which confuses the offsets.
    # The default, newline = None, casuses replacement.  Errors = 'replace'
    # causes bad UTF-8 chars to be replaced with '?'.
    return open(filename,'r', errors='strict', newline='').read()

########################################################################

def relpath(p1, p2) :
    return os.path.relpath(p1, p2)

def joinpath(*p1) :
    return os.path.join(*p1)

def dirname(p1) :
    return os.path.dirname(p1)

def realpath(p1) :
    return os.path.realpath(p1)



import re
import markdown
import xml.dom.minidom
from urllib.parse import urlparse
import http.client
import os
import tempfile
import urllib.request
import urllib.parse
import urllib.error


start_single_quote_re = re.compile("(^|\s|\")'")
start_double_quote_re = re.compile("(^|\s|'|`)\"")
end_double_quote_re = re.compile("\"(,|\.|\s|$)")


def inline_html_latex(text):
    out = text
    # most of them to support smarty extensions
    if re.search(r'&ldquo;.*?&rdquo;', text, flags=re.DOTALL):
        out = out.replace('&ldquo;', '\enquote{').replace('&rdquo;', '}')
    # replace certain html element with they LaTeX eqivarent
    if re.search(r'&lsquo;.*?&rsquo;', text, flags=re.DOTALL):
        out = out.replace('&lsquo;', '\enquote{').replace('&rsquo;', '}')
    if re.search(r'&ldquo;.*?&ldquo;', text, flags=re.DOTALL):
        # sometimes is processing like this
        out = out.replace('&ldquo;', '\enquote{', 1).replace('&ldquo;', '}', 1)
    if re.search(r'&laquo;.*?&raquo;', text, flags=re.DOTALL):
        out = out.replace('&laquo;', '\enquote{').replace('&raquo;', '}')
    out = out.replace("...", "\dots")
    out = out.replace("&hellip;", "\dots")
    out = out.replace("&ndash;", "--")
    out = out.replace("&mdash;", "---")
    # replace '\|' as we should already processed the tables and do not need in LaTeX
    out = out.replace("\|", '|')
    return out


def unescape_html_entities(text):
    out = text.replace('&amp;', '&')
    out = out.replace('&lt;', '<')
    out = out.replace('&gt;', '>')
    out = out.replace('&quot;', '"')
    return out


def escape_latex_entities(text):
    """Escape latex reserved characters."""
    out = text
    out = unescape_html_entities(out)
    out = out.replace('%', '\\%')
    out = out.replace('&', '\\&')
    out = out.replace('#', '\\#')
    out = start_single_quote_re.sub('\g<1>`', out)
    out = start_double_quote_re.sub('\g<1>``', out)
    out = end_double_quote_re.sub("''\g<1>", out)
    # people should escape these themselves as it conflicts with maths
    # out = out.replace('{', '\\{')
    # out = out.replace('}', '\\}')
    # do not do '$' here because it is dealt with by convert_maths
    # out = out.replace('$', '\\$')
    return out


def unescape_latex_entities(text):
    """Limit ourselves as this is only used for maths stuff."""
    out = text
    out = out.replace('\\&', '&')
    return out


def makeExtension(configs=None):
    return LaTeXExtension(configs=configs)


class LaTeXExtension(markdown.Extension):
    def __init__(self, configs=None):
        self.reset()

    def extendMarkdown(self, md, md_globals):
        self.md = md

        # remove escape pattern -- \\(.*) -- as this messes up any embedded
        # math and we don't need to escape stuff any more for html
        # for key, pat in self.md.inlinePatterns.items():
        #     if pat.pattern == markdown.inlinepatterns.ESCAPE_RE:
        #         self.md.inlinePatterns.pop(key)
        #         break

        #footnote_extension = FootnoteExtension()
        #footnote_extension.extendMarkdown(md, md_globals)

        latex_tp = LaTeXTreeProcessor()
        math_pp = MathTextPostProcessor()
        table_pp = TableTextPostProcessor()
        image_pp = ImageTextPostProcessor()
        link_pp = LinkTextPostProcessor()
        unescape_html_pp = UnescapeHtmlTextPostProcessor()

        md.treeprocessors['latex'] = latex_tp
        md.postprocessors['unescape_html'] = unescape_html_pp
        md.postprocessors['math'] = math_pp
        md.postprocessors['image'] = image_pp
        md.postprocessors['table'] = table_pp
        md.postprocessors['link'] = link_pp

    def reset(self):
        pass


class LaTeXTreeProcessor(markdown.treeprocessors.Treeprocessor):
    def run(self, doc):
        """Walk the dom converting relevant nodes to text nodes with relevant
        content."""
        latex_text = self.tolatex(doc)

        doc.clear()
        latex_node = markdown.util.etree.Element('div')
        latex_node.text = latex_text
        doc.append(latex_node)

    def tolatex(self, ournode):
        buffer = ""
        subcontent = ""

        if ournode.text:
            subcontent += escape_latex_entities(ournode.text)
        if list(ournode):
            for child in list(ournode):
                subcontent += self.tolatex(child)

        if ournode.tag == 'h1':
            if "OVERVIEW" in subcontent:
                buffer += """\\begin{Large}\\textbf{%s}\\end{Large}\\newline\\newline""".strip() % subcontent
            else: 
                buffer += """\\begin{large}\\textbf{%s}\\end{large}""".strip() % subcontent
        elif ournode.tag == 'h2':
            buffer += """\\begin{large}\\textbf{%s}\\end{large}""".strip() % subcontent
        elif ournode.tag == 'h3':
            buffer += """\\begin{normalsize}\\textbf{%s}\\end{normalsize}""".strip() % subcontent
        elif ournode.tag == 'h4':
            buffer += """\\begin{small}\\textbf{%s}\\end{small}""".strip() % subcontent
        elif ournode.tag == 'hr':
            buffer += '\\noindent\makebox[\linewidth]{\\rule{\linewidth}{0.4pt}}'
        elif ournode.tag == 'ul':
            # no need for leading \n as one will be provided by li
            buffer += """
\\begin{itemize}%s
\\end{itemize}
""" % subcontent
        elif ournode.tag == 'ol':
            buffer += """
\\begin{enumerate}"""
            if 'start' in ournode.attrib.keys():
                start = int(ournode.attrib['start'])-1
                buffer += "\setcounter{enumi}{"+str(start)+"}"
            # no need for leading \n as one will be provided by li
            buffer += """
%s
\\end{enumerate}
""" % subcontent
        elif ournode.tag == 'li':
            buffer += """
  \\item %s""" % subcontent.strip()
        elif ournode.tag == 'blockquote':
            # use quotation rather than quote as quotation can support multiple
            # paragraphs
            buffer += """
\\begin{quotation}
%s
\\end{quotation}
""" % subcontent.strip()
        # ignore 'code' when inside pre tags
        # (mkdn produces <pre><code></code></pre>)
        elif (ournode.tag == 'pre' or
                             # TODO: Take a look here
              (ournode.tag == 'pre' and ournode.parentNode.tag != 'pre')):
            buffer += """
\\begin{verbatim}
%s
\\end{verbatim}
""" % subcontent.strip()
        elif ournode.tag == 'q':
            buffer += "`%s'" % subcontent.strip()
        elif ournode.tag == 'p':
            contents = subcontent.split(" ")
            buffer += '\n'
            for content in contents:
                if content.startswith("http"):
                    buffer += ' \\href{%s}{%s}' % (content,content)
                else:
                    buffer += ' %s' % content.strip()
            buffer += '\n'
        # Footnote processor inserts all of the footnote in a sup tag
        elif ournode.tag == 'sup':
            buffer += '\\footnote{%s}' % subcontent.strip()
        elif ournode.tag == 'strong':
            buffer += '\\textbf{%s}' % subcontent.strip()
        elif ournode.tag == 'em':
            buffer += '\\emph{%s}' % subcontent.strip()
        # Keep table strcuture. TableTextPostProcessor will take care.
        elif ournode.tag == 'table':
            buffer += '\n\n<table>%s</table>\n\n' % subcontent
        elif ournode.tag == 'thead':
            buffer += '<thead>%s</thead>' % subcontent
        elif ournode.tag == 'tbody':
            buffer += '<tbody>%s</tbody>' % subcontent
        elif ournode.tag == 'tr':
            buffer += '<tr>%s</tr>' % subcontent
        elif ournode.tag == 'th':
            buffer += '<th>%s</th>' % subcontent
        elif ournode.tag == 'td':
            buffer += '<td>%s</td>' % subcontent
        elif ournode.tag == 'img':
            buffer += '<img src=\"%s\" alt=\"%s\" />' % (ournode.get('src'),
                                                         ournode.get('alt'))
        elif ournode.tag == 'a':
            buffer += '<a href=\"%s\">%s</a>' % (escape_latex_entities(ournode.get('href')),
                                                 subcontent)
        else:
            buffer = subcontent

        if ournode.tail:
            buffer += escape_latex_entities(ournode.tail)

        return buffer


class UnescapeHtmlTextPostProcessor(markdown.postprocessors.Postprocessor):

    def run(self, text):
        return unescape_html_entities(inline_html_latex(text))

# ========================= MATHS =================================


class MathTextPostProcessor(markdown.postprocessors.Postprocessor):

    def run(self, instr):
        """Convert all math sections in {text} whether latex, asciimathml or
        latexmathml formatted to latex.
        This assumes you are using $ for inline math and $$ for blocks as your
        mathematics delimiter (*not* the standard asciimathml or latexmathml
        delimiter).
        """
        def repl_1(matchobj):
            text = unescape_latex_entities(matchobj.group(1))
            return '\\[%s\\]' % text

        def repl_2(matchobj):
            text = unescape_latex_entities(matchobj.group(1))
            return '\\(%s\\)' % text

        # This $$x=3$$ is block math
        pat = re.compile('\$\$([^\$]*)\$\$')
        out = pat.sub(repl_1, instr)
        # This $x=3$ is inline math
        pat2 = re.compile('\$([^\$]*)\$')
        out = pat2.sub(repl_2, out)
        # some extras due to asciimathml
        out = out.replace('\\lt', '<')
        out = out.replace(' * ', ' \\cdot ')
        out = out.replace('\\del', '\\partial')
        return out


# ========================= TABLES =================================

class TableTextPostProcessor(markdown.postprocessors.Postprocessor):

    def run(self, instr):
        """This is not very sophisticated and for it to work it is expected
        that:
            1. tables to be in a section on their own (that is at least one
            blank line above and below)
            2. no nesting of tables
        """
        converter = Table2Latex()
        new_blocks = []

        for block in instr.split('\n\n'):
            stripped = block.strip()
            # <table catches modified verions (e.g. <table class="..">
            if stripped.startswith('<table') and stripped.endswith('</table>'):
                latex_table = converter.convert(stripped).strip()
                new_blocks.append(latex_table)
            else:
                new_blocks.append(block)
        return '\n\n'.join(new_blocks)


class Table2Latex:
    """
    Convert html tables to Latex.
    TODO: escape latex entities.
    """

    def colformat(self):
        # centre align everything by default
        out = '|l' * self.maxcols + '|'
        return out

    def get_text(self, element):
        if element.nodeType == element.TEXT_NODE:
            return escape_latex_entities(element.data)
        result = ''
        if element.childNodes:
            for child in element.childNodes:
                text = self.get_text(child)
                if text.strip() != '':
                    result += text
        return result

    def process_cell(self, element):
        # works on both td and th
        colspan = 1
        subcontent = self.get_text(element)
        buffer = ""

        if element.tagName == 'th':
            subcontent = '\\textbf{%s}' % subcontent
        if element.hasAttribute('colspan'):
            colspan = int(element.getAttribute('colspan'))
            buffer += ' \multicolumn{%s}{|c|}{%s}' % (colspan, subcontent)
        # we don't support rowspan because:
        #   1. it needs an extra latex package \usepackage{multirow}
        #   2. it requires us to mess around with the alignment tags in
        #   subsequent rows (i.e. suppose the first col in row A is rowspan 2
        #   then in row B in the latex we will need a leading &)
        # if element.hasAttribute('rowspan'):
        #     rowspan = int(element.getAttribute('rowspan'))
        #     buffer += ' \multirow{%s}{|c|}{%s}' % (rowspan, subcontent)
        else:
            buffer += ' %s' % subcontent

        notLast = (element.nextSibling.nextSibling and
                   element.nextSibling.nextSibling.nodeType ==
                   element.ELEMENT_NODE and
                   element.nextSibling.nextSibling.tagName in ['td', 'th'])

        if notLast:
            buffer += ' &'

        self.numcols += colspan
        return buffer

    def tolatex(self, element):
        if element.nodeType == element.TEXT_NODE:
            return ""

        buffer = ""
        subcontent = ""
        if element.childNodes:
            for child in element.childNodes:
                text = self.tolatex(child)
                if text.strip() != "":
                    subcontent += text
        subcontent = subcontent.strip()

        if element.tagName == 'thead':
            buffer += subcontent

        elif element.tagName == 'tr':
            self.maxcols = max(self.numcols, self.maxcols)
            self.numcols = 0
            buffer += '\n\\hline\n%s \\\\' % subcontent

        elif element.tagName == 'td' or element.tagName == 'th':
            buffer = self.process_cell(element)
        else:
            buffer += subcontent
        return buffer

    def convert(self, instr):
        self.numcols = 0
        self.maxcols = 0
        dom = xml.dom.minidom.parseString(instr)
        core = self.tolatex(dom.documentElement)

        captionElements = dom.documentElement.getElementsByTagName('caption')
        caption = ''
        if captionElements:
            caption = self.get_text(captionElements[0])

        colformatting = self.colformat()
        table_latex = \
            """
            \\begin{table}[h]
            \\begin{tabular}{%s}
            %s
            \\hline
            \\end{tabular}
            \\\\[5pt]
            \\caption{%s}
            \\end{table}
            """ % (colformatting, core, caption)
        return table_latex


# ========================= IMAGES =================================

class ImageTextPostProcessor(markdown.postprocessors.Postprocessor):

    def run(self, instr):
        """Process all img tags
        Similar to process_tables this is not very sophisticated and for it
        to work it is expected that img tags are put in a section of their own
        (that is separated by at least one blank line above and below).
        """
        converter = Img2Latex()
        new_blocks = []
        for block in instr.split("\n\n"):
            stripped = block.strip()
            # <table catches modified verions (e.g. <table class="..">
            if stripped.startswith('<img'):
                latex_img = converter.convert(stripped).strip()
                new_blocks.append(latex_img)
            else:
                new_blocks.append(block)
        return '\n\n'.join(new_blocks)


class Img2Latex(object):
    def convert(self, instr):
        dom = xml.dom.minidom.parseString(instr)
        img = dom.documentElement
        src = img.getAttribute('src')

        if urlparse(src).scheme != '':
            src_urlparse = urlparse(src)
            conn = http.client.HTTPConnection(src_urlparse.netloc)
            conn.request('HEAD', src_urlparse.path)
            response = conn.getresponse()
            conn.close()
            if response.status == 200:
                filename = os.path.join(tempfile.mkdtemp(), src.split('/')[-1])
                urllib.request.urlretrieve(src, filename)
                src = filename

        alt = img.getAttribute('alt')
        # Using graphicx and ajustbox package for *max width*
        out = \
            """
            \\begin{figure}[H]
            \\centering
            \\includegraphics[max width=\\linewidth]{%s}
            \\caption{%s}
            \\end{figure}
            """ % (src, alt)
        return out


# ========================== LINKS =================================

class LinkTextPostProcessor(markdown.postprocessors.Postprocessor):

    def run(self, instr):
        # Process all hyperlinks
        converter = Link2Latex()
        new_blocks = []
        for block in instr.split("\n\n"):
            stripped = block.strip()
            match = re.findall(r'<a[^>]*>[^<]+</a>', stripped)
            # <table catches modified verions (e.g. <table class="..">
            if match:
                latex_link = stripped
                # replace multiple <a> occurrences individually
                for occurrence in match:
                    latex_link = re.sub(r'<a[^>]*>([^<]+)</a>',
                                        converter.convert(occurrence).strip(),
                                        latex_link, count=1)
                new_blocks.append(latex_link)
            else:
                new_blocks.append(block)
        return '\n\n'.join(new_blocks)


class Link2Latex(object):
    def convert(self, instr):
        dom = xml.dom.minidom.parseString(instr)
        link = dom.documentElement
        href = link.getAttribute('href')

        desc = re.search(r'>([^<]+)', instr)
        out = \
            """
            \\\href{%s}{%s}
            """ % (href, desc.group(0)[1:])
        return out


"""
========================= FOOTNOTES =================================
LaTeX footnote support.
Implemented via modification of original markdown approach (place footnote
definition in footnote market <sup> as opposed to putting a reference link).
"""


class FootnoteExtension (markdown.Extension):
    DEF_RE = re.compile(r"(\ ?\ ?\ ?)\[\^([^\]]*)\]:\s*(.*)")
    SHORT_USE_RE = re.compile(r"\[\^([^\]]*)\]", re.M)  # [^a]

    def __init__(self, configs=None):
        self.reset()

    def extendMarkdown(self, md, md_globals):
        self.md = md

        # Stateless extensions do not need to be registered
        md.registerExtension(self)

        # Insert a preprocessor before ReferencePreprocessor
        #index = md.preprocessors.index(md_globals['REFERENCE_PREPROCESSOR'])
        #preprocessor = FootnotePreprocessor(self)
        #preprocessor.md = md
        #md.preprocessors.insert(index, preprocessor)
        md.preprocessors.add('footnotes', FootnotePreprocessor(self), '_begin')

        # Insert an inline pattern before ImageReferencePattern
        FOOTNOTE_RE = r"\[\^([^\]]*)\]"  # blah blah [^1] blah
        #index = md.inlinePatterns.index(md_globals['IMAGE_REFERENCE_PATTERN'])
        #md.inlinePatterns.insert(index, FootnotePattern(FOOTNOTE_RE, self))
        md.inlinePatterns.add('footnotes', FootnotePattern(FOOTNOTE_RE, self),
                              '_begin')

    def reset(self):
        self.used_footnotes = {}
        self.footnotes = {}

    def setFootnote(self, id, text):
        self.footnotes[id] = text


class FootnotePreprocessor:
    def __init__(self, footnotes):
        self.footnotes = footnotes

    def run(self, lines):
        lines = self._handleFootnoteDefinitions(lines)

        # Make a hash of all footnote marks in the text so that we
        # know in what order they are supposed to appear.  (This
        # function call doesn't really substitute anything - it's just
        # a way to get a callback for each occurence.

        text = "\n".join(lines)
        self.footnotes.SHORT_USE_RE.sub(self.recordFootnoteUse, text)

        return text.split("\n")

    def recordFootnoteUse(self, match):
        id = match.group(1)
        id = id.strip()
        nextNum = len(list(self.footnotes.used_footnotes.keys())) + 1
        self.footnotes.used_footnotes[id] = nextNum

    def _handleFootnoteDefinitions(self, lines):
        """Recursively finds all footnote definitions in the lines.
            @param lines: a list of lines of text
            @returns: a string representing the text with footnote
                      definitions removed """

        i, id, footnote = self._findFootnoteDefinition(lines)

        if id:

            plain = lines[:i]

            detabbed, theRest = lines[i + 1:].strip()

            self.footnotes.setFootnote(id,
                                       footnote + "\n"
                                       + "\n".join(detabbed))

            more_plain = self._handleFootnoteDefinitions(theRest)
            return plain + [""] + more_plain

        else:
            return lines

    def _findFootnoteDefinition(self, lines):
        """Finds the first line of a footnote definition.
            @param lines: a list of lines of text
            @returns: the index of the line containing a footnote definition.
        """

        counter = 0
        for line in lines:
            m = self.footnotes.DEF_RE.match(line)
            if m:
                return counter, m.group(2), m.group(3)
            counter += 1
        return counter, None, None


class FootnotePattern(markdown.inlinepatterns.Pattern):

    def __init__(self, pattern, footnotes):
        markdown.inlinepatterns.Pattern.__init__(self, pattern)
        self.footnotes = footnotes

    def handleMatch(self, m, doc):
        sup = doc.createElement('sup')
        id = m.group(2)
        # stick the footnote text in the sup
        self.footnotes.md._processSection(sup,
                                          self.footnotes.footnotes[id].split("\n"))
        return sup


from ntpath import join
from operator import index
import os
import re
import subprocess
import time

from lxml import etree
import markdown
from ecldoc.Utils import read_file, write_to_file
from ecldoc.Utils import joinpath, relpath, dirname

from ecldoc.Constants import TEMPLATE_DIR
from ecldoc.markdown2latex import LaTeXExtension

TEX_TEMPLATE_DIR = joinpath(TEMPLATE_DIR, 'tex')


import jinja2
latex_jinja_env = jinja2.Environment(
    block_start_string = '\\BLOCK{',
    block_end_string = '}',
    variable_start_string = '\\VAR{',
    variable_end_string = '}',
    comment_start_string = '\\#{',
    comment_end_string = '}',
    line_statement_prefix = '%%',
    line_comment_prefix = '%#',
    trim_blocks = True,
    autoescape = False,
    loader = jinja2.FileSystemLoader(os.path.abspath('/'))
)

LATEX_SUBS = (
    (re.compile(r'\\'), r'\\textbackslash '),
    (re.compile(r'([{}_#%&$])'), r'\\\1'),
    (re.compile(r'~'), r'\~{}'),
    (re.compile(r'\^'), r'\^{}'),
    (re.compile(r'"'), r"''"),
    (re.compile(r'\.\.\.+'), r'\\ldots'),
)

def escape_tex(value):
    newval = value
    for pattern, replacement in LATEX_SUBS:
        newval = pattern.sub(replacement, newval)
    return newval

latex_jinja_env.filters['escape_tex'] = escape_tex


from ecldoc.parseDoc import getTags
from ecldoc.Taglets import taglets
from .tagTEX import tag_renders

class ParseTEX(object) :
    '''
    Main class to generate TEX & PDF Documentation for given ecl file
    from its XML Repr
    '''
    def __init__(self, generator, ecl_file) :
        self.xml_file = joinpath(generator.xml_root, ecl_file + '.xml')
        self.tex_file = joinpath(generator.tex_root, ecl_file + '.tex')

        tex_relpath = relpath(self.tex_file, generator.tex_path)
        self.dirname = dirname(tex_relpath)

        self.template = generator.content_template
        self.options = generator.options

        os.makedirs(dirname(self.tex_file), exist_ok=True)

    def parse(self) :
        root = etree.parse(self.xml_file).getroot()
        src = root.find('Source')
        self.src = src
        self.doc = src.find('Documentation')

        for child in root.iter() :
            attribs = child.attrib
            if 'target' in attribs :
                attribs['target'] = re.sub(r'\$\$_ECLDOC-FORM_\$\$', joinpath('tex', 'root'), attribs['target'])
                attribs['target'] = re.sub(r'\.xml$', '.tex', attribs['target'])

        name = src.attrib['name'].split('.')
        self.parseSource()

        render = self.template.render(name=name, src=src, defn_tree=self.defn_tree, up=('toc:'+self.dirname))
        write_to_file(self.tex_file, render)

    def docstring(self) :
        text = ''
        if self.doc is not None :
            content = self.doc.find('firstline')
            if content is not None :
                text = content.text
        return text

    def parseSource(self) :
        '''
        Append all Definitions under source tag into defn_tree
        '''
        self.defn_tree = []
        for defn in self.src.findall('Definition') :
            defn_dict = self.parseDefinition(defn)
            self.defn_tree.append(defn_dict)

    def parseDefinition(self, defn) :
        '''
        Generate Dict representation of Definition for appending into defn_tree
        '''
        sign = defn.find('Signature')
        doc = self.parseDocs(defn)

        defn_dict = { 'tag' : defn, 'sign' : sign, 'doc' : doc, 'defns' : [] }

        ### Append all Children of current Definition to defn_tree
        for childdefn in defn.findall('Definition') :
            child_dict = self.parseDefinition(childdefn)
            defn_dict['defns'].append(child_dict)

        return defn_dict

    def parseDocs(self, defn) :
        renders = {}
        tags = getTags(defn.find('Documentation'))
        always = ['param', 'field', 'return', 'parent', 'content', 'firstline']
        common_tags = list(set(tags.keys()) | set(always))
        for tag in common_tags :
            if tag not in taglets or tag not in tag_renders :
                if tag in ['content', 'firstline'] :
                    assert False, 'Taglet not found for required tags (content, firstline)'
                if 'generaltag' in tag_renders :
                    render = tag_renders['generaltag'](
                                taglets['generaltag'](doc=tags[tag], defn=defn, tagname=tag))
                    renders[tag] = render
                continue
            render = tag_renders[tag](taglets[tag](doc=tags[tag], defn=defn, tagname=tag))
            renders[tag] = render

        renders['inherit'] = tag_renders['inherit'](defn.attrib['inherittype'])

        return renders

class GenTEX(object) :
    '''
    Generate TEX & PDF Documentation for all ecl files from XML Format
    '''
    def __init__(self, input_root, output_root, ecl_file_tree, options) :
        self.input_root = input_root
        self.output_root = output_root

        self.tex_path = joinpath(output_root, 'tex')
        self.tex_root = joinpath(output_root, 'tex', 'root')
        self.xml_root = joinpath(output_root, 'xml')
        os.makedirs(self.tex_root, exist_ok=True)

        self.template_dir = TEX_TEMPLATE_DIR
        self.content_template = latex_jinja_env.get_template(joinpath(self.template_dir, 'content.tpl.tex'))
        self.toc_template = latex_jinja_env.get_template(joinpath(self.template_dir, 'toc.tpl.tex'))
        self.index_template = latex_jinja_env.get_template(joinpath(self.template_dir, 'index.tpl.tex'))

        self.ecl_file_tree = ecl_file_tree
        self.options = options

    def gen(self, key, node, content_root) :
        '''
        Recursively parse source tree dictionary.
        If current_node is file : parse file using parseTEX
        Else If : current_node is directory : recurse and generate pkg.toc.tex for that dir
                                              (optionally generate bundle info if present)
        :param key: string | the name of current_node in path tree
        :param node: dict | the parent tree of current_node
        :param content_root: string | real path to current node in tex doc dir
        '''
        current_node = node[key]
        if type(current_node) != dict:
            if key == 'bundle.ecl' :
                return
            parser = ParseTEX(self, current_node)
            parser.parse()
            file = { 'name' : key, 'type' : 'file', 'doc' : parser.docstring() }
            file['target'] = relpath(parser.tex_file, self.tex_path)
            file['label'] = parser.src.attrib['name']
            return file
        else :
            os.makedirs(content_root, exist_ok=True)
            render_path = joinpath(content_root, 'pkg.toc.tex')
            temptoc_render_path = joinpath(content_root, 'pkg.tmp.tex')
            index_render_path = joinpath(content_root, 'index.tex')

            tex_relpath = relpath(content_root, self.tex_path)
            target_relpath = relpath(render_path, self.tex_path)

            file = { 'name' : key, 'type': 'dir', 'doc' : '' , 'target' : target_relpath, 'label' : tex_relpath }

            bundle = None
            if 'bundle.ecl' in current_node :
                bundle_xml_path = joinpath(self.xml_root, dirname(current_node['bundle.ecl']), 'bundle.xml')
                bundle = etree.parse(bundle_xml_path).getroot()
                file['type'] = 'bundle'

            childfiles = []
            child_keys = sorted(current_node.keys(), key=str.lower)
            for chkey in child_keys :
                child_root = joinpath(content_root, chkey)
                child_dict = self.gen(chkey, current_node, child_root)
                if child_dict is not None : childfiles.append(child_dict)

            childfiles = sorted(childfiles, key=lambda x : x['type'], reverse=True)

            render = self.toc_template.render(name=key, files=childfiles, bundle=bundle,
                                            label=tex_relpath, up=dirname(tex_relpath))

            render = render.replace("<p>", '').replace("</p>", '').replace('<div>', '').replace('</div>', '')
            if key == "root":
                end_of_toc = "\\end{tabularx}\n\n"
                render_temp = render.split(end_of_toc)
                readme_md_path = joinpath(self.input_root, 'README.md')
                readme_data_in_markdown = read_file(readme_md_path)
                readme_data_in_markdown = "# OVERVIEW\n" + readme_data_in_markdown
                md = markdown.Markdown()
                mkdn2latex = LaTeXExtension()
                mkdn2latex.extendMarkdown(md, markdown.__dict__)
                readmeLatexContent = md.convert(readme_data_in_markdown).replace("</div>", "").replace("<div>", "")
                print(readme_data_in_markdown)
                print(readmeLatexContent)
                render = """{render_temp_start}{end_of_toc}{readme}\n\n{render_temp_end}""".format(
                    render_temp_start=render_temp[0],
                    render_temp_end=render_temp[1], 
                    end_of_toc=end_of_toc,
                    readme= readmeLatexContent
                ).strip()
            write_to_file(render_path, render)

            render = self.toc_template.render(name=key,
                                            files=[x for x in childfiles if x['type'] == 'file'],
                                            bundle=bundle, label=tex_relpath, up="")
            
            render = render.replace("<p>", '').replace("</p>", '').replace('<div>', '').replace('</div>', '')
            write_to_file(temptoc_render_path, render)

            # Render index.pdf
            start_path = relpath(temptoc_render_path, self.tex_path)
            render = self.index_template.render(root=start_path)

            render = render.replace("<p>", '').replace("</p>", '').replace('<div>', '').replace('</div>', '')
            write_to_file(index_render_path, render)
            subprocess.run(['pdflatex --shell-escape ' +
                            '-output-directory ' + relpath(content_root, self.tex_path) + ' ' +
                            relpath(index_render_path, self.tex_path)],
                            cwd=self.tex_path, shell=True)

            return file

    def run(self) :
        '''
        Main Function called by ECLDOC
        '''
        print("\nGenerating PDF Documentation ... ")
        self.gen('root', self.ecl_file_tree, self.tex_root)

        render_path = joinpath(self.tex_root, 'pkg.toc.tex')
        start_path = relpath(render_path, self.tex_path)
        render = self.index_template.render(root=start_path)
        write_to_file(joinpath(self.tex_path, 'index.tex'), render)

        subprocess.run(['pdflatex --shell-escape index.tex'], cwd=self.tex_path, shell=True)

        render = render.replace("<p>", '').replace("</p>", '').replace('<div>', '').replace('</div>', '')
        write_to_file(joinpath(self.tex_path, 'index.tex'), render)

        subprocess.run(['pdflatex index.tex'], cwd=self.tex_path, shell=True)